/*
 * Copyright (c), Lawrence Foard, 1992
 *
 * Developed under contract by Virtual World Information systems.
 * Questions and suggestions about this package may be directed to:
 * lfoard@turing.org, entropy@world.std.com
 * 
 * Those without e-mail access can call the VWIS Linux support BBS
 * at 1-508-793-9568 and send e-mail to entropy,
 * or send snail mail to:
 *
 * Virtual World Information Systems
 * 217 Park Ave.
 * Box #108
 * Worcester MA 01609
 *
 * This code may be used under the terms of the GNU license.
 * Other licensing agreements may be negotiated with VWIS.
 *
 */
 
#include <stdio.h>
#include <fcntl.h>
#include "head.h"

/* Xenix a.out header */

struct x_out
 {
  unsigned short magic;
  unsigned short extend;
  int text;
  int data;
  int bss;
  int symbol;
  int reloc;
  int entry;
  char cpu;
  char relsym;
  unsigned short renv;
  /* extended header */
  int text_relocation;
  int data_relocation;
  int text_base;
  int data_base;
  int stack_size;
  int seg_table;
  int seg_table_size;
  int mach_table;
  int mach_table_size;
  char mach_t_type;
  char page_size;
  char os_type;
  char os_version;
  unsigned short entry_seg;
  unsigned short reserved;
 };

/* segment header */
struct seg 
 {
  unsigned short type;
  unsigned short attr;
  unsigned short number;
  char align;
  char reserved;
  int position;
  int real_size;
  int virtual_size;
  int relocation;
  unsigned short name_offset;
  unsigned short sres;
  int lres;
 };

struct x_out *get_header(int fd)
 {
  static struct x_out x_out;
  lseek(fd,0,0);
  read(fd,&x_out,sizeof(x_out));
  return(&x_out);
 }

struct seg *get_seg(int fd,struct x_out *out,int segn)
 {
  static struct seg seg;
  if (segn*sizeof(struct seg)>=out->seg_table_size)
   return(NULL);
  lseek(fd,out->seg_table+segn*sizeof(struct seg),0);
  read(fd,&seg,sizeof(seg));
  return(&seg);
 }

void print_seg(FILE *debug,struct seg *seg)
 {
  if (!debug)
   return;
  fprintf(debug,"segment type %x\n",seg->type);
  fprintf(debug,"segment attributes %x\n",seg->attr);
  fprintf(debug,"segment number %x\n",seg->number);
  fprintf(debug,"segment alignment %x\n",seg->align);
  fprintf(debug,"file position %x\n",seg->position);
  fprintf(debug,"real size %x\n",seg->real_size);
  fprintf(debug,"virtual size %x\n",seg->virtual_size);
  fprintf(debug,"relocation %x\n",seg->relocation);
  fprintf(debug,"name offset %x\n",seg->name_offset);
  fprintf(debug,"\n");
 }

void print_dat(FILE *debug,struct x_out *out)
 {
  if (!debug)
   return;
  fprintf(debug,"Magic %x\n",out->magic);
  fprintf(debug,"Extension length %x\n",out->extend);
  fprintf(debug,"Text size %x\n",out->text);
  fprintf(debug,"Data size %x\n",out->data);
  fprintf(debug,"Bss size %x\n",out->bss);
  fprintf(debug,"Symbol table size %x\n",out->symbol);
  fprintf(debug,"Relocation length %x\n",out->reloc);
  fprintf(debug,"Entry point %x\n",out->entry);
  fprintf(debug,"Relocation format %x\n",out->relsym);
  fprintf(debug,"Environment %x\n",out->renv);
  fprintf(debug,"Segment table pos %x\n",out->seg_table);
  fprintf(debug,"Segment table size %x\n",out->seg_table_size);
  fprintf(debug,"\n");
 }

extern char end_restorer,start_restorer;

PROCESS *load_file(char *file,FILE *debug,int stack_size,int heap_size)
 {
  int fd,a;
  struct x_out *out;
  struct seg *seg;
  PROCESS *process=(PROCESS *) malloc(sizeof(PROCESS));
  process->debug=debug;
  /* try to open the program */
  if (-1==(fd=open(file,O_RDONLY)))
   {
    perror(file);
    exit(1);
   }
  out=get_header(fd);
  if (process->debug)
   print_dat(process->debug,out);
  process->prog_size=0;
  /* figure out size of area to initially allocate */
  for(a=0;seg=get_seg(fd,out,a);a++)
   if ((seg->virtual_size+seg->relocation)>process->prog_size)
    process->prog_size=seg->virtual_size+seg->relocation;
  process->seg_break=process->prog_size;
  /* reserve room for the heap and stack */
  process->prog_size+=heap_size+stack_size+4100 /* limit granularity */;
  a=sbrk(process->prog_size+4096);
  a=((a>>12)+1)<<12; /* force page alignment */
  process->prog_area=(char *)a;
  /* load segments */
  for(a=0;seg=get_seg(fd,out,a);a++)
   {
    FILE *fo;
    if (process->debug)
     print_seg(process->debug,seg);
    lseek(fd,seg->position,0);
    if ((seg->type!=1) && (seg->type!=2))
     {
      fprintf(stderr,"Ignoring segment type %x\n",seg->type);
      continue;
     }
    read(fd,process->prog_area+seg->relocation,seg->real_size);
   }
  process->user_stack=process->prog_size-100;
  /* copy signal stack restorer here */
  process->restorer=(process->user_stack-=&end_restorer-&start_restorer);
  memcpy(&process->prog_area[process->restorer],&start_restorer,
         &end_restorer-&start_restorer);
  process->user_stack-=10;
  process->user_stack=(1+process->user_stack/sizeof(int))*sizeof(int);
  process->map=NULL;
  if (process->debug)
   {
    fprintf(process->debug,"Total process size=%d\n",process->prog_size);
    fflush(process->debug);
   }
  close(fd);
  return process;
 }

/* add file name mapping */
void add_map(PROCESS *process,char *prefix,char *replace)
 {
  REMAP *tmp=(REMAP *) malloc(sizeof(REMAP));
  strcpy(tmp->prefix=(char *) malloc(strlen(prefix)+1),prefix);
  strcpy(tmp->replace=(char *) malloc(strlen(replace)+1),replace);
  tmp->next=process->map;
  process->map=tmp;
 }

/* set process arguments and environment */
void set_argvnc(PROCESS *process,char *argv[],char *envp[])
 {
  int slen;
  int string;
  int *stack;
  int a;
  if (process->debug)
   {
    fprintf(process->debug,"Argument assignment start\n");
    fflush(process->debug);
   }
  for(slen=process->nargs=0;argv[process->nargs];process->nargs++)
   slen+=strlen(argv[process->nargs])+1;
  for(a=0;envp[a];a++)
   slen+=strlen(envp[a])+1;
  slen=(1+(slen/sizeof(int)))*sizeof(int);
  string=process->user_stack-slen;
  process->user_stack-=slen+sizeof(int)*(process->nargs+a)+sizeof(int)*10;
  stack=(int *)&process->prog_area[process->user_stack];
  for(a=0;argv[a];a++)
   {
    stack[a+1]=string;
    strcpy(&process->prog_area[string],argv[a]);
    string+=strlen(argv[a])+1;
   }
  stack[a+1]=0;
  stack[0]=process->nargs;
  stack=stack+a+1;
  for(a=0;envp[a];a++)
   {
    stack[a+1]=string;
    strcpy(&process->prog_area[string],envp[a]);
    string+=strlen(envp[a])+1;
   }
  stack[a+1]=0;
  if (process->debug)
   {
    fprintf(process->debug,"Argument assignment complete\n");
    fflush(process->debug);
   }
 }

